import java.util.TreeMap;

import java.util.LinkedList;
import java.util.List;
import java.math.BigDecimal;
import java.math.MathContext;

public class Tri
{

  private final CartPoint top;
  private final CartPoint left;
  private final CartPoint right;

  public Tri(CartPoint top, CartPoint left, CartPoint right)
  {
    this.top = top;
    this.left = left;
    this.right = right;

  }

  public String toString()
  {
    return String.format("Tri{%s %s %s}", top.toString(), left.toString(), right.toString());
  }

  public List<CartPoint> getPoints()
  {
    LinkedList<CartPoint> lst = new LinkedList<>();
    lst.add(top);
    lst.add(left);
    lst.add(right);

    return lst;

  }

  public CartPoint getCenter()
  {
    CartPoint c = new CartPoint("0,0,0");
    c = c.add(top);
    c = c.add(left);
    c = c.add(right);

    c = c.multi(1.0/3.0);
    return c;
  }

  public double getCenterDist(CartPoint check)
  {
    return getCenter().getDist(check).doubleValue();
  }

	public BigDecimal getArea()
	{
    BigDecimal a = top.getDist(left);
    BigDecimal b = left.getDist(right);
    BigDecimal c = right.getDist(top);

    BigDecimal s = a.add(b).add(c).divide(BigDecimal.valueOf(2), new MathContext(200));

    //double sub = s * (s - a) * (s - b) * (s - c);
    BigDecimal sub = s
      .multiply(s.subtract(a))
      .multiply(s.subtract(b))
      .multiply(s.subtract(c));

    BigDecimal area = sub.sqrt(new MathContext(200));


    //System.out.println(String.format("area: %.9f %.9f %.9f (%.9f) - %.9f",a,b,c,sub,area));

    return area;

	}

  public boolean pointInTri(CartPoint point)
  {
    return pointInTri(point, false);
  }
  public boolean pointInTri(CartPoint point, boolean verbose)
  {
    Tri rad_t = this.radConv();

    point = point.getRad().getCart();

    BigDecimal area = rad_t.getArea();

    BigDecimal a_sum = BigDecimal.ZERO;

    a_sum = a_sum.add(new Tri(point, rad_t.left, rad_t.top).getArea());
    a_sum = a_sum.add(new Tri(point, rad_t.left, rad_t.right).getArea());
    a_sum = a_sum.add(new Tri(point, rad_t.top, rad_t.right).getArea());

    if (verbose)
    {

      System.out.println("Point in: " + a_sum + " " + area);
    }

    BigDecimal area_diff = a_sum.subtract(area);
    BigDecimal area_ratio = area_diff.abs().divide(area, new MathContext(200));

    if (verbose)
    {
      System.out.println("Area diff: " + area_diff);
      System.out.println("Area ratio: " + area_ratio);
    }
    BigDecimal limit = new BigDecimal("1.9");
    return (area_ratio.compareTo(limit) < 0);


  }

  public double pointInTriRatio(CartPoint point, boolean verbose)
  {
    Tri rad_t = this.radConv();

    point = point.getRad().getCart();

    BigDecimal area = rad_t.getArea();

    BigDecimal a_sum = BigDecimal.ZERO;

    a_sum = a_sum.add(new Tri(point, rad_t.left, rad_t.top).getArea());
    a_sum = a_sum.add(new Tri(point, rad_t.left, rad_t.right).getArea());
    a_sum = a_sum.add(new Tri(point, rad_t.top, rad_t.right).getArea());

    if (verbose)
    {

      System.out.println("Point in: " + a_sum + " " + area);
    }

    BigDecimal area_diff = a_sum.subtract(area);
    BigDecimal area_ratio = area_diff.abs().divide(area, new MathContext(200));

    return area_ratio.doubleValue();


  }


  /*public boolean pointInTriC(CartPoint point)
  {
    RadPoint p = point.getRad();

    RadPoint p0 = top.getRad();
    RadPoint p1 = left.getRad();
    RadPoint p2 = right.getRad();

    double s = (p0.X() - p2.X()) * (p.Y() - p2.Y()) - (p0.Y() - p2.Y()) * (p.X() - p2.X());

    double t = (p1.X() - p0.X()) * (p.Y() - p0.Y()) - (p1.Y() - p0.Y()) * (p.X() - p0.X());

    //if ((s < 0) != (t < 0) && s != 0 && t != 0)
    //    return false;

    double d = (p2.X() - p1.X()) * (p.Y() - p1.Y()) - (p2.Y() - p1.Y()) * (p.X() - p1.X());
		//System.out.println("std " + s + " " + t + " " + d);
		//if (Math.abs(d) < 1e-10) return true;
    return d == 0 || (d < 0) == (s + t <= 0);
    

  }*/

  /*public boolean maybePointIn(CartPoint p)
  {
    CartPoint pr = p.getRad().getCart();
    Tri r = radConv();

    double m=r.pointMaxDist(pr);

    double side=0.0;

    side = Math.max(r.top.getDist(r.right), side);
    side = Math.max(r.top.getDist(r.left), side);
    side = Math.max(r.left.getDist(r.right), side);
    side = side * 1.05;

    if (m<= side) return true;

    return false;
  }*/

  public Tri radConv()
  {
    return new Tri( top.getRad().getCart(), left.getRad().getCart(), right.getRad().getCart());

  }

  public static LinkedList<Integer> getPath(CartPoint target, Tri major, int steps)
  {
    LinkedList<Integer> path = new LinkedList<>();
    
    if (steps == 0) return path;

    TreeMap<Integer, Tri> subs = getTriangles(major);

    //TreeMap<Double, Integer> best = new TreeMap<>();
    TreeMap<PathOption, Integer> best = new TreeMap<>();
    for(int idx : subs.keySet())
    {
      //if (subs.get(idx).maybePointIn(target))
      if (subs.get(idx).pointInTri(target, false))
      {
        //double d = subs.get(idx).getCenterDist(target);
        double d = subs.get(idx).pointInTriRatio(target,false);

        best.put(
          new PathOption(d, getPath(target, subs.get(idx), steps -1)), 
          idx);
      }
    }

    if (best.size() == 0) return path;

    int best_idx = best.firstEntry().getValue();

    path.add(best_idx);

    path.addAll(best.firstKey().path);

    //path.addAll(getPath(target, subs.get(best_idx), steps-1));

    return path;
  }

  public static class PathOption implements Comparable<PathOption>
  {
    final double center_dist;
    final LinkedList<Integer> path;
    public PathOption(double center_dist, List<Integer> path)
    {
      this.center_dist = center_dist;
      this.path = new LinkedList<>();
      this.path.addAll(path);

    }
    public int compareTo(PathOption o)
    {
      if (path.size() > o.path.size()) return -1;
      if (path.size() < o.path.size()) return 1;

      if (center_dist < o.center_dist) return -1;
      if (center_dist > o.center_dist) return 1;

      return 0;


    }

  }


  /**
   * Get the 64 sub triangles of a given triangle
   */
  public static TreeMap<Integer, Tri> getTriangles(Tri major)
  {
    CartPoint[][] points = new CartPoint[9][9];

    points[0][0] = major.top;

    CartPoint v_down = major.top.getDirection(major.left).multi(1.0/8.0);
    CartPoint v_right = major.left.getDirection(major.right).multi(1.0/8.0);

    for(int row=1; row<=8; row++)
    {
      for(int c=0; c<=8; c++)
      {
        CartPoint p = major.top.add(v_down.multi(row)).add(v_right.multi(c));
        points[row][c]=p;
      }
    }

    int row_items = 1;
    int t_idx=1;
    TreeMap<Integer, Tri> subs=new TreeMap<>();
    for(int row=0; row<8; row++)
    { 
      for(int col=0; col<row_items; col++)
      {
        if (col%2 == 0)
        { //upright
          Tri t = new Tri(points[row][col/2], points[row+1][col/2], points[row+1][col/2+1]);
          subs.put(t_idx, t);
        }
        else
        { //inverted
          Tri t = new Tri(points[row+1][col/2+1], points[row][col/2+1], points[row][col/2]);
          subs.put(t_idx, t);
        }
        t_idx++;
      }
      row_items+=2;
    }

    BigDecimal area_sum = BigDecimal.ZERO;

    for(int idx : subs.keySet())
    {
      for(CartPoint p : subs.get(idx).getPoints())
      {
        /*if (!major.maybePointIn(p))
        {
          System.out.println("slop point in " + idx);

        }*/
		    if (!major.pointInTri(p))
        {
          System.out.println("slop tri point in " + idx);
          major.pointInTri(p,true);
        }
      }
      area_sum = area_sum.add( subs.get(idx).getArea() );

    }
    //System.out.println("areas: " + major.getArea() + " " + area_sum);

    return subs;

  }


}
